import requests

from ....sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ..abstract import AbstractCommand


class WorldDeleteCommand(AbstractCommand):
    async def execute(self, request: WorldDeleteRequest) -> None:
        requests.delete(url=f"http://{self.options.tld}/world/{request.id}", timeout=self.options.timeout)



import asyncio
import logging
import uuid
from asyncio import Queue

from pydantic import BaseModel

from ....client.dao import DaoClient
from ....sdk.contracts.dtos.entities.entity import Entity
from ....sdk.contracts.dtos.tiles.address import Address
from ....sdk.contracts.dtos.tiles.tile import Tile
from ....sdk.contracts.errors.server.factory import FactoryError
from ....sdk.contracts.types.entity import EntityType
from ....sdk.contracts.types.tile import TileType

logger = logging.getLogger()


class AbstractEntityFactory(BaseModel):
    daoclient: DaoClient

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    async def producer(ids: set[str], queue: Queue):
        for local_id in ids:
            await queue.put(local_id)

    async def generate(self, address: Address) -> None:
        # get entities ids for the tile
        local_tile: Tile = await self.daoclient.get(address=address)

        if len(local_tile.ids) > 0:
            msg: str = f"entity generation can not occur on a tile with pre-existing entities, {address}"
            raise FactoryError(msg)

        # Review types now
        if local_tile.tile_type == TileType.GRASS:
            new_entity: Entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.GRASS)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

        elif local_tile.tile_type == TileType.FOREST:
            new_entity: Entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.TREE)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

            new_entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.MYCELIUM)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

        elif local_tile.tile_type == TileType.OCEAN:
            new_entity: Entity = Entity(id=str(uuid.uuid4()), entity_type=EntityType.FISH)
            address_entity: Address = Address.model_validate({**address.model_dump(), "entity_id": new_entity.id})
            await self.daoclient.post(address=address_entity, document=new_entity)
            local_tile.ids.add(new_entity.id)

        patch: dict = {"ids": list(local_tile.ids)}
        # print(f"tile id {local_tile}, got: {patch}")
        await self.daoclient.patch(address=address, document=patch)

    ### entity agent logic
    async def entity_mycelium(self, entity: Entity):
        pass

    async def entity_grass(self, entity: Entity):
        pass

    async def entity_fish(self, entity: Entity):
        pass

    async def entity_tree(self, entity: Entity):
        pass

    ###

    async def grow_entities(self, address: Address):
        # get entities ids for the tile
        local_tile: Tile = await self.daoclient.get(address=address)

        async def entity_producer(queue: Queue):
            for entity_id in local_tile.ids:
                await queue.put(entity_id)

        async def step_one():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_entity_id: str = await queue.get()

                    # TODO: Process entity
                    address_entity: Address = Address.model_validate(
                        {**address.model_dump(), "entity_id": local_entity_id}
                    )
                    entity: Entity = await self.daoclient.get(address=address_entity)
                    # print(wrapped_entity.model_dump())

                    if entity.entity_type == EntityType.MYCELIUM:
                        await self.entity_mycelium(entity=entity)
                    # elif wrapped_entity.data.entity_type == EntityType.MUSHROOM:
                    #     pass
                    elif entity.entity_type == EntityType.GRASS:
                        await self.entity_grass(entity=entity)
                    elif entity.entity_type == EntityType.FISH:
                        await self.entity_fish(entity=entity)
                    elif entity.entity_type == EntityType.TREE:
                        await self.entity_tree(entity=entity)
                    # elif wrapped_entity.data.entity_type == EntityType.FRUIT:
                    #     pass

                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(entity_producer(queue), consumer(queue))

        await step_one()



import logging
import os

import click
import uvicorn
from fastapi import FastAPI

from ...factories.world import world
from ..common.routers import metrics
from .routers import world, worlds

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


# pylint: disable=too-many-arguments,no-value-for-parameter
@click.command()
@click.option("--hostname", type=click.STRING, default="0.0.0.0", help="host address to bind to")
@click.option("--port", type=click.INT, default=8000, help="port to bind to")
@click.option("--log-level", type=click.STRING, default="INFO", help="log level (INFO, DEBUG, WARNING, ERROR, FATAL)")
@click.option("--sleep-time", type=click.FLOAT, default=1.0, help="api call sleep time (seconds)")
@click.option("--timeout", type=click.FLOAT, default=5.0, help="api call timeout (seconds)")
@click.option("--retries", type=click.INT, default=5, help="api call retries (integer)")
@click.option("--url", type=click.STRING, is_flag=False, default=None, help="darkness api url, format: '(host):(port)'")
def main(
    hostname: str, port: int, log_level: str, sleep_time: float, timeout: float, retries: int, url: str | None
) -> None:
    logger.setLevel(logging.getLevelName(log_level))

    # Setup server runtime environment variables
    if url:
        os.environ["DARKNESS_TLD"] = url
    else:
        os.environ["DARKNESS_TLD"] = f"{hostname}:{port}"
    os.environ["DARKNESS_SERVICE_SLEEP_TIME"] = str(sleep_time)
    os.environ["DARKNESS_SERVICE_RETRY_COUNT"] = str(retries)
    os.environ["DARKNESS_SERVICE_TIMEOUT"] = str(timeout)

    logger.info("[Main] starting")
    app = FastAPI()

    logger.debug("[Main] adding metrics routes")
    app.include_router(metrics.router)

    logger.debug("[Main] adding world routes")
    app.include_router(world.router)

    logger.debug("[Main] adding worlds routes")
    app.include_router(worlds.router)

    logger.info("[Main] online")

    uvicorn.run(app, host=hostname, port=port, log_level=logging.getLevelName(log_level))


if __name__ == "__main__":
    main()



import json

from pydantic import BaseModel
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.results import DeleteResult, InsertOneResult, UpdateResult

from ..sdk.common.utils import address_type, get_document_id_from_address
from ..sdk.contracts.dtos.entities.entity import Entity
from ..sdk.contracts.dtos.tiles.address import Address
from ..sdk.contracts.dtos.tiles.chunk import Chunk
from ..sdk.contracts.dtos.tiles.tile import Tile
from ..sdk.contracts.dtos.tiles.world import World
from ..sdk.contracts.types.dao_document import DaoDocumentType


def get_mongodb(hostname: str, port: int, database: str) -> Database:
    return MongoClient(hostname, port)[database]


class DaoClient(BaseModel):
    database: Database
    collections: dict[DaoDocumentType, Collection] = {}

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # our collections will match our document types
        for member in DaoDocumentType:
            # Create a collection for each document type
            self.collections[DaoDocumentType[member.name]] = self.database[member.value]

            # We add a single column index on the business logic layer `id` for our lookups:
            # db.<collection>.createIndex( { <field>: <sortOrder> } )
            self.collections[DaoDocumentType[member.name]].create_index({"id": 1})

    class Config:
        arbitrary_types_allowed = True

    # Single document delete
    async def delete(self, address: Address) -> DeleteResult:
        doc_type = address_type(address)
        document_id: str = get_document_id_from_address(address=address, doc_type=doc_type)
        return self.collections[doc_type].delete_one({"id": document_id})

    # single document get with typing!
    async def get(self, address: Address) -> World | Chunk | Tile | Entity:
        doc_type: DaoDocumentType = address_type(address=address)
        document_id: str = get_document_id_from_address(address=address, doc_type=doc_type)
        result: dict | None = self.collections[doc_type].find_one({"id": document_id})
        if result is None:
            raise Exception("no document found")
        if doc_type == DaoDocumentType.ENTITY:
            return Entity.model_validate(result)
        elif doc_type == DaoDocumentType.TILE:
            return Tile.model_validate(result)
        elif doc_type == DaoDocumentType.CHUNK:
            return Chunk.model_validate(result)
        elif doc_type == DaoDocumentType.WORLD:
            return World.model_validate(result)
        raise Exception("invalid document type")

    async def get_all(self, doc_type: DaoDocumentType) -> list[World] | list[Chunk] | list[Tile] | list[Entity]:
        results = self.collections[doc_type].find()
        documents: list[World] | list[Chunk] | list[Tile] | list[Entity] = []
        for result in results:
            if doc_type == DaoDocumentType.ENTITY:
                documents.append(Entity.model_validate(result))
            elif doc_type == DaoDocumentType.TILE:
                documents.append(Tile.model_validate(result))
            elif doc_type == DaoDocumentType.CHUNK:
                documents.append(Chunk.model_validate(result))
            elif doc_type == DaoDocumentType.WORLD:
                documents.append(World.model_validate(result))
            else:
                raise Exception("invalid document type")
        return documents

    async def get_multi(
        self, addresses: list[Address], doc_type: DaoDocumentType
    ) -> list[World] | list[Chunk] | list[Tile] | list[Entity]:
        doc_ids: list = []
        for address in addresses:
            document_id: str = get_document_id_from_address(address=address, doc_type=doc_type)
            doc_ids.append(document_id)
        results = self.collections[doc_type].find({"id": {"$in": doc_ids}})
        documents: list[World] | list[Chunk] | list[Tile] | list[Entity] = []
        for result in results:
            if doc_type == DaoDocumentType.ENTITY:
                documents.append(Entity.model_validate(result))
            elif doc_type == DaoDocumentType.TILE:
                documents.append(Tile.model_validate(result))
            elif doc_type == DaoDocumentType.CHUNK:
                documents.append(Chunk.model_validate(result))
            elif doc_type == DaoDocumentType.WORLD:
                documents.append(World.model_validate(result))
            else:
                raise Exception("invalid document type")
        return documents

    async def post(self, address: Address, document: World | Chunk | Tile | Entity) -> InsertOneResult:
        doc_type: DaoDocumentType = address_type(address=address)
        return self.collections[doc_type].insert_one(json.loads(document.model_dump_json()))

    async def patch(self, address: Address, document: dict) -> UpdateResult:
        # document CAN NOT CONTAIN set variables.  It MUST be serialized before this call, so help you
        doc_type: DaoDocumentType = address_type(address=address)
        document_id: str = get_document_id_from_address(address=address, doc_type=doc_type)
        return self.collections[doc_type].update_one(filter={"id": document_id}, update={"$set": document})



import asyncio
import json
import logging
import re
import secrets
import uuid
from abc import abstractmethod
from asyncio import Queue

from pydantic import BaseModel

from ....client.dao import DaoClient
from ....sdk.contracts.dtos.entities.entity import Entity
from ....sdk.contracts.dtos.tiles.address import Address
from ....sdk.contracts.dtos.tiles.chunk import Chunk
from ....sdk.contracts.dtos.tiles.tile import Tile
from ....sdk.contracts.dtos.tiles.world import World
from ....sdk.contracts.dtos.window import Window
from ....sdk.contracts.types.connection import TileConnectionType
from ....sdk.contracts.types.dao_document import DaoDocumentType
from ....sdk.contracts.types.tile import TileType

logger = logging.getLogger()


class AbstractChunkFactory(BaseModel):
    daoclient: DaoClient

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    async def producer(ids: set[str], queue: Queue):
        for local_id in ids:
            await queue.put(local_id)

    @abstractmethod
    async def create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> str:
        """ """

    async def mutate_tile(self, address: Address, mutate: float, tile_type: TileType) -> None:
        if secrets.randbelow(100) <= mutate:
            # then we spawn the tile type
            await self.daoclient.patch(address=address, document={"tile_type": tile_type})

    async def convert_tile(self, address: Address, source: TileType, target: TileType) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)
        # check
        if target_tile.tile_type == source:
            # update
            await self.daoclient.patch(address=address, document={"tile_type": target})

    async def adjecent_liquids(self, address: Address) -> list[TileType]:
        return await self.adjecent_to(address=address, types=[TileType.OCEAN, TileType.WATER])

    async def adjecent_to(self, address: Address, types: list[TileType] | None) -> list[TileType]:
        adjecent_targets: list[TileType] = []

        target_tile: Tile = await self.daoclient.get(address=address)

        adjecent_tile_addresses: list[Address] = []
        for _, adjecent_id in target_tile.next.items():
            adjecent_tile_addresses.append(Address.model_validate({**address.model_dump(), "tile_id": adjecent_id}))

        adjecent_tiles: list[Tile] = await self.daoclient.get_multi(
            addresses=adjecent_tile_addresses, doc_type=DaoDocumentType.TILE
        )
        for adjecent_tile in adjecent_tiles:
            if adjecent_tile.tile_type in types:
                adjecent_targets.append(adjecent_tile.tile_type)

        return list(set(adjecent_targets))

    async def grow_tile(self, address: Address) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)
        # dirt -> grass
        if target_tile.tile_type == TileType.DIRT:
            adjecent_liquids: list[TileType] = await self.adjecent_liquids(address=address)
            if TileType.WATER in adjecent_liquids and TileType.OCEAN not in adjecent_liquids:
                await self.daoclient.patch(address=address, document={"tile_type": TileType.GRASS})

        # grass+water (no dirt/ocean) -> forest
        if target_tile.tile_type == TileType.GRASS:
            neighbors: list[TileType] = await self.adjecent_to(
                address=address,
                types=[TileType.WATER, TileType.GRASS, TileType.OCEAN],
            )

            # no forests grow by oceans
            if TileType.OCEAN in neighbors:
                return

            if len(neighbors) > 1:
                # next to more than one kind (grass/water)
                await self.daoclient.patch(address=address, document={"tile_type": TileType.FOREST})

        # TODO: grass+(dirt)

    async def brackish_tile(self, address: Address) -> None:
        # Convert inner Ocean to Water Tiles

        # See if we are next to another ocean tile
        neighbors: list[TileType] = await self.adjecent_to(address=address, types=[TileType.OCEAN])
        if len(neighbors) < 1:
            await self.convert_tile(address=address, source=TileType.OCEAN, target=TileType.WATER)

        # TODO: are we an isolated ocean body? if so then we are water
        #
        # # 1. For each next that is liquid (recusively), can we get to the edge?
        # #     # 1. Are we ocean? - if no then return False
        # #     # 2. Are we edge? - if yes then return True
        # #     # 3. then check my nexts
        # #     # 4. Return False (not ocean)

    async def erode_tile(self, address: Address) -> None:
        # get
        target_tile: Tile = await self.daoclient.get(address=address)

        # shore erosion
        if target_tile.tile_type not in [TileType.UNKNOWN, TileType.OCEAN, TileType.WATER, TileType.SHORE]:
            adjecent_liquids: list[TileType] = await self.adjecent_liquids(address=address)
            # Apply erosion - rocks and be left by oceans, everything else becomes shore
            if TileType.OCEAN in adjecent_liquids:
                if target_tile.tile_type not in (TileType.ROCK, TileType.SHORE):
                    await self.daoclient.patch(address=address, document={"tile_type": TileType.SHORE})

    async def gen_geo_bind(self, address: Address, conn_dir: TileConnectionType, target_tile_id: str) -> None:
        # TODO: move to native mongodb to make this efficient ..
        doc: World | Chunk | Tile | Entity = await self.daoclient.get(address=address)
        doc.next[conn_dir] = target_tile_id
        await self.daoclient.patch(address=address, document=json.loads(doc.model_dump_json()))

    async def generate_ocean_block(self, address: Address, window: Window):
        tile_map: dict[str, str] = {}

        async def flat_producer(window: Window, queue: Queue):
            range_x_min: int = window.min.x - 1
            range_x_max: int = window.max.x
            range_y_min: int = window.min.x - 1
            range_y_max: int = window.max.y

            for x in range(range_x_min, range_x_max):
                for y in range(range_y_min, range_y_max):
                    local_x = x + 1
                    local_y = y + 1
                    tile_id: str = f"tile_{local_x}_{local_y}"
                    await queue.put(tile_id)

        # 1. fill a blank nXm area with ocean
        async def step_one():
            # get container
            async def consumer(queue: Queue):
                chunk: Chunk = await self.daoclient.get(address=address)

                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    tile_map[local_tile_id] = str(uuid.uuid4())
                    local_tile: Tile = Tile(id=tile_map[local_tile_id], tile_type=TileType.OCEAN)
                    address_tile: Address = Address.model_validate({**address.model_dump(), "tile_id": local_tile.id})

                    # create tile
                    await self.daoclient.post(address=address_tile, document=local_tile)
                    # msg: str = f"({local_tile_id}) brought into existence as {TileType.OCEAN}"
                    # logger.debug(msg)

                    # update
                    chunk.ids.add(tile_map[local_tile_id])
                    queue.task_done()

                # Update the chunk --
                # put -- store chunk update (tile addition)
                document = {"ids": list(chunk.ids)}
                # print(document)
                await self.daoclient.patch(address=address, document=document)

            queue = asyncio.Queue()
            await asyncio.gather(flat_producer(window, queue), consumer(queue))

        await step_one()

        # set origin tile on chunk
        await self.daoclient.patch(address=address, document={"origin": tile_map["tile_1_1"]})

        # 2. Connect everything together
        async def step_two():
            pattern: re.Pattern = re.compile("^tile_([a-zA-Z0-9-]+)_([a-zA-Z0-9-]+)$")

            async def consumer(queue):
                while not queue.empty():
                    _local_tile_id: str = await queue.get()
                    local_tile_id = tile_map[_local_tile_id]

                    match = pattern.search(_local_tile_id)
                    # tile_id: str = f"tile_{local_x}_{local_y}"
                    local_x: int = int(match.group(1))
                    local_y: int = int(match.group(2))

                    # msg: str = f"binding ({local_tile_id}) physically to peers"
                    # logger.debug(msg)

                    # Bind LEFT
                    _target_tile_id: str = f"tile_{local_x - 1}_{local_y}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.LEFT,
                        )

                    # Bind RIGHT
                    _target_tile_id: str = f"tile_{local_x + 1}_{local_y}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.RIGHT,
                        )

                    # Bind UP
                    _target_tile_id: str = f"tile_{local_x}_{local_y - 1}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.UP,
                        )

                    # Bind DOWN
                    _target_tile_id: str = f"tile_{local_x}_{local_y + 1}"
                    if _target_tile_id in tile_map:
                        target_tile_id = tile_map[_target_tile_id]
                        address_tile: Address = Address.model_validate(
                            {**address.model_dump(), "tile_id": local_tile_id}
                        )
                        await self.gen_geo_bind(
                            address=address_tile,
                            target_tile_id=target_tile_id,
                            conn_dir=TileConnectionType.DOWN,
                        )

                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(flat_producer(window, queue), consumer(queue))

        await step_two()



import requests

from ....sdk.contracts.dtos.sdk.requests.chunk.get import ChunkGetRequest
from ....sdk.contracts.dtos.sdk.responses.chunk.get import ChunkGetResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ..abstract import AbstractCommand


class ChunkGetCommand(AbstractCommand):
    async def execute(self, request: ChunkGetRequest) -> ChunkGetResponse:
        response: requests.Response = requests.get(
            url=f"http://{self.options.tld}/world/{request.world_id}/chunk/{request.chunk_id}",
            timeout=self.options.timeout,
            params={"full": request.full},
        )
        return Response[ChunkGetResponse].model_validate(response.json()).data



# Excluding server from package exports (explicitly)
from .client.commands.abstract import AbstractCommand
from .client.commands.chunk.create import ChunkCreateCommand
from .client.commands.chunk.delete import ChunkDeleteCommand
from .client.commands.chunk.get import ChunkGetCommand
from .client.commands.metrics.health.get import HealthGetCommand
from .client.commands.world.create import WorldCreateCommand
from .client.commands.world.delete import WorldDeleteCommand
from .client.commands.world.get import WorldGetCommand
from .client.commands.worlds.get import WorldsGetCommand
from .client.dao import DaoClient
from .client.state import StateClient
from .sdk.common.config.environment import (
    demand_env_var,
    demand_env_var_as_bool,
    demand_env_var_as_float,
    demand_env_var_as_int,
    get_env_var,
)
from .sdk.contracts.dtos.coordinate import Coordinate
from .sdk.contracts.dtos.entities.abstract import AbstractEntity
from .sdk.contracts.dtos.entities.entity import Entity
from .sdk.contracts.dtos.sdk.command_options import CommandOptions
from .sdk.contracts.dtos.sdk.request_status_codes import RequestStatusCodes
from .sdk.contracts.dtos.sdk.requests.chunk.create import ChunkCreateRequest
from .sdk.contracts.dtos.sdk.requests.chunk.delete import ChunkDeleteRequest
from .sdk.contracts.dtos.sdk.requests.chunk.get import ChunkGetRequest
from .sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from .sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from .sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from .sdk.contracts.dtos.sdk.responses.chunk.create import ChunkCreateResponse
from .sdk.contracts.dtos.sdk.responses.chunk.delete import ChunkDeleteResponse
from .sdk.contracts.dtos.sdk.responses.chunk.get import ChunkGetResponse
from .sdk.contracts.dtos.sdk.responses.response import Response
from .sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from .sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from .sdk.contracts.dtos.sdk.responses.worlds.get import WorldsGetResponse
from .sdk.contracts.dtos.sdk.wrapped_request import WrappedRequest
from .sdk.contracts.dtos.tiles.abtract import AbstractTile
from .sdk.contracts.dtos.tiles.chunk import Chunk
from .sdk.contracts.dtos.tiles.tile import Tile
from .sdk.contracts.dtos.tiles.world import World
from .sdk.contracts.dtos.window import Window
from .sdk.contracts.errors.common.environment_variable_not_found import EnvironmentVariableNotFoundError
from .sdk.contracts.errors.sdk.exceeded_retry_count import ExceededRetryCountError
from .sdk.contracts.errors.sdk.request_failure import RequestFailureError
from .sdk.contracts.errors.sdk.unknown_verb import UnknownVerbError
from .sdk.contracts.errors.server.dao.conflict import DaoConflictError
from .sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from .sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError
from .sdk.contracts.errors.server.factory import FactoryError
from .sdk.contracts.errors.server.service import ServiceError
from .sdk.contracts.types.connection import TileConnectionType
from .sdk.contracts.types.dao_document import DaoDocumentType
from .sdk.contracts.types.entity import EntityType
from .sdk.contracts.types.sdk.request_verb import RequestVerbType
from .sdk.contracts.types.tile import TileType



import uuid

from pydantic import BaseModel

from ....client.dao import DaoClient
from ....sdk.contracts.dtos.tiles.address import Address
from ....sdk.contracts.dtos.tiles.world import World


class WorldFactory(BaseModel):
    daoclient: DaoClient

    class Config:
        arbitrary_types_allowed = True

    async def create(self, name: str | None = None) -> str:
        if name is None:
            name = "darkness"
        world: World = World(id=str(uuid.uuid4()), name=name)
        address_world: Address = Address.model_validate({"world_id": world.id})

        await self.daoclient.post(address=address_world, document=world)
        return world.id



from pathlib import Path

from pydantic import BaseModel


class Address(BaseModel):
    world_id: str | None = None
    chunk_id: str | None = None
    tile_id: str | None = None
    entity_id: str | None = None

    def resolve(self) -> Path:
        path: Path = Path()
        if self.world_id is not None:
            path = path / "worlds" / self.world_id
        if self.chunk_id is not None:
            path = path / "chunks" / self.chunk_id
        if self.tile_id is not None:
            path = path / "tiles" / self.tile_id
        if self.entity_id is not None:
            path = path / "entities" / self.entity_id
        path = path / "metadata.json"
        return path



from ..contracts.dtos.tiles.address import Address
from ..contracts.types.dao_document import DaoDocumentType


def address_type(address: Address) -> DaoDocumentType:
    if address.entity_id is not None:
        return DaoDocumentType.ENTITY
    elif address.tile_id is not None:
        return DaoDocumentType.TILE
    elif address.chunk_id is not None:
        return DaoDocumentType.CHUNK
    elif address.world_id is not None:
        return DaoDocumentType.WORLD

    raise Exception("Unknown address type")


def get_document_id_from_address(address: Address, doc_type: DaoDocumentType | None = None) -> str:
    if doc_type is None:
        doc_type = address_type(address)

    if doc_type == DaoDocumentType.ENTITY:
        return address.entity_id
    elif doc_type == DaoDocumentType.TILE:
        return address.tile_id
    elif doc_type == DaoDocumentType.CHUNK:
        return address.chunk_id
    elif doc_type == DaoDocumentType.WORLD:
        return address.world_id
    raise Exception("Unknown address type")



from ..sdk.contracts.dtos.sdk.command_options import CommandOptions
from ..sdk.contracts.dtos.sdk.requests.chunk.create import ChunkCreateRequest
from ..sdk.contracts.dtos.sdk.requests.chunk.delete import ChunkDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.chunk.get import ChunkGetRequest
from ..sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ..sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ..sdk.contracts.dtos.sdk.responses.chunk.create import ChunkCreateResponse
from ..sdk.contracts.dtos.sdk.responses.chunk.get import ChunkGetResponse
from ..sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from ..sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from ..sdk.contracts.dtos.sdk.responses.worlds.get import WorldsGetResponse
from ..sdk.contracts.dtos.tiles.chunk import Chunk
from ..sdk.contracts.dtos.tiles.world import World
from ..sdk.contracts.types.tile import TileType
from .commands.chunk.create import ChunkCreateCommand
from .commands.chunk.delete import ChunkDeleteCommand
from .commands.chunk.get import ChunkGetCommand
from .commands.metrics.health.get import HealthGetCommand
from .commands.world.create import WorldCreateCommand
from .commands.world.delete import WorldDeleteCommand
from .commands.world.get import WorldGetCommand
from .commands.worlds.get import WorldsGetCommand


class StateClient:
    # metrics/health
    health_get_command: HealthGetCommand

    # chunk
    chunk_create_command: ChunkCreateCommand
    chunk_delete_command: ChunkDeleteCommand
    chunk_get_command: ChunkGetCommand

    # world
    world_create_command: WorldCreateCommand
    world_get_command: WorldGetCommand
    world_delete_command: WorldDeleteCommand

    # worlds
    worlds_get_command: WorldsGetCommand

    def __init__(self, options: CommandOptions | None = None):
        command_dict: dict = {}
        if options:
            command_dict["options"] = options

        ### metrics
        self.health_get_command = HealthGetCommand.model_validate(command_dict)

        ### chunk
        self.chunk_create_command = ChunkCreateCommand.model_validate(command_dict)
        self.chunk_delete_command = ChunkDeleteCommand.model_validate(command_dict)
        self.chunk_get_command = ChunkGetCommand.model_validate(command_dict)

        # world
        self.world_create_command = WorldCreateCommand.model_validate(command_dict)
        self.world_get_command = WorldGetCommand.model_validate(command_dict)
        self.world_delete_command = WorldDeleteCommand.model_validate(command_dict)

        # worlds
        self.worlds_get_command = WorldsGetCommand.model_validate(command_dict)

    # metrics

    async def health_get(self) -> dict:
        """
        Gets the server health.

        Returns
        -------
        health: dict
            The server health (true or false).
        """

        return await self.health_get_command.execute()

    # chunk

    async def chunk_create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> str:
        request: ChunkCreateRequest = ChunkCreateRequest(
            world_id=world_id, name=name, dimensions=dimensions, biome=biome
        )
        response: ChunkCreateResponse = await self.chunk_create_command.execute(request=request)
        return response.id

    async def chunk_get(self, world_id: str, chunk_id: str, full: bool = False) -> Chunk:
        request: ChunkGetRequest = ChunkGetRequest(world_id=world_id, chunk_id=chunk_id, full=full)
        response: ChunkGetResponse = await self.chunk_get_command.execute(request=request)
        return response.chunk

    async def chunk_delete(self, world_id: str, chunk_id: str) -> None:
        request: ChunkDeleteRequest = ChunkDeleteRequest(world_id=world_id, chunk_id=chunk_id)
        await self.chunk_delete_command.execute(request=request)

    # world

    async def world_create(self, name: str | None) -> str:
        request: WorldCreateRequest = WorldCreateRequest(name=name)
        response: WorldCreateResponse = await self.world_create_command.execute(request)
        return response.id

    async def world_delete(self, world_id: str) -> None:
        request: WorldDeleteRequest = WorldDeleteRequest(world_id=world_id)
        await self.world_delete_command.execute(request=request)

    async def world_get(self, world_id: str, full: bool) -> World:
        request: WorldGetRequest = WorldGetRequest(world_id=world_id, full=full)
        response: WorldGetResponse = await self.world_get_command.execute(request)
        return response.world

    # worlds

    async def worlds_get(self) -> list[World]:
        response: WorldsGetResponse = await self.worlds_get_command.execute()
        return response.worlds



import requests

from ....sdk.contracts.dtos.sdk.requests.chunk.create import ChunkCreateRequest
from ....sdk.contracts.dtos.sdk.responses.chunk.create import ChunkCreateResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ..abstract import AbstractCommand


class ChunkCreateCommand(AbstractCommand):
    async def execute(self, request: ChunkCreateRequest) -> ChunkCreateResponse:
        response: requests.Response = requests.post(
            url=f"http://{self.options.tld}/world/{request.world_id}/chunk",
            timeout=self.options.timeout,
            json=request.model_dump(),
        )
        return Response[ChunkCreateResponse].model_validate(response.json()).data



import asyncio
import logging
import uuid
from asyncio import Queue

from ....sdk.contracts.dtos.coordinate import Coordinate
from ....sdk.contracts.dtos.tiles.address import Address
from ....sdk.contracts.dtos.tiles.chunk import Chunk
from ....sdk.contracts.dtos.tiles.world import World
from ....sdk.contracts.dtos.window import Window
from ....sdk.contracts.types.tile import TileType
from .abstract import AbstractChunkFactory

logger = logging.getLogger()


class FlatChunkFactory(AbstractChunkFactory):
    async def terrain_generate(self, address: Address, chunk: Chunk) -> None:
        # Lets shovel in some biome default or dirt tiles!
        async def step_one():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    address_tile: Address = Address.model_validate({**address.model_dump(), "tile_id": local_tile_id})
                    # Mutate tiles to biome default (or dirt)
                    await self.mutate_tile(
                        address=address_tile,
                        mutate=90,  # percentage of 100%
                        tile_type=(chunk.biome if chunk.biome else TileType.DIRT),
                    )
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatChunkFactory.producer(ids=chunk.ids, queue=queue), consumer(queue))

        await step_one()

        # Add Eradicates (rocks)
        async def step_two():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    address_tile: Address = Address.model_validate({**address.model_dump(), "tile_id": local_tile_id})
                    await self.mutate_tile(
                        address=address_tile,
                        mutate=0.5,  # 0.5% change (very low)
                        tile_type=TileType.ROCK,
                    )
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(self.producer(ids=chunk.ids, queue=queue), consumer(queue))

        await step_two()

        # Convert inner Ocean to Water Tiles
        async def step_three():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    address_tile: Address = Address.model_validate({**address.model_dump(), "tile_id": local_tile_id})
                    # Convert inner Ocean to Water Tiles
                    await self.brackish_tile(address=address_tile)
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(self.producer(ids=chunk.ids, queue=queue), consumer(queue))

        await step_three()

        # Erode Tiles (to make shore)
        async def step_four():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    address_tile: Address = Address.model_validate({**address.model_dump(), "tile_id": local_tile_id})
                    await self.erode_tile(address=address_tile)
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(self.producer(ids=chunk.ids, queue=queue), consumer(queue))

        await step_four()

    async def create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> Chunk:
        if name is None:
            name = "roshar"

        address_world: Address = Address.model_validate({"world_id": world_id})

        # 1. blank, named chunk
        chunk: Chunk = Chunk(id=str(uuid.uuid4()), name=name, dimensions=dimensions, biome=biome)
        address_chunk: Address = Address.model_validate({**address_world.model_dump(), "chunk_id": chunk.id})
        await self.daoclient.post(address=address_chunk, document=chunk)

        # update world metadata
        world: World = await self.daoclient.get(address=address_world)
        world.ids.add(chunk.id)
        await self.daoclient.patch(address=address_world, document={"ids": list(world.ids)})

        # Define the maximum size
        max_x, max_y = dimensions

        # address["chunk_id"] = chunk.id
        # Generate an empty 2D block of ocean
        window: Window = Window(min=Coordinate(x=1, y=1), max=Coordinate(x=max_x, y=max_y))
        await self.generate_ocean_block(address=address_chunk, window=window)

        chunk: Chunk = await self.daoclient.get(address=address_chunk)

        # Apply our terrain generation
        await self.terrain_generate(address=address_chunk, chunk=chunk)

        # Apply a quantum time
        await self.quantum(world_id=world_id, chunk=chunk)

        # get final state and return
        return await self.daoclient.get(address=address_chunk)

    async def quantum(self, world_id: str, chunk: Chunk) -> None:
        address_chunk: Address = Address.model_validate({"world_id": world_id, "chunk_id": chunk.id})

        # Grow Tiles
        async def step_five():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    address_tile: Address = Address.model_validate(
                        {**address_chunk.model_dump(), "tile_id": local_tile_id}
                    )
                    await self.grow_tile(address=address_tile)
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(self.producer(ids=chunk.ids, queue=queue), consumer(queue))

        await step_five()



from ...types.tile import TileType
from .abtract import AbstractTile
from .tile import Tile


class Chunk(AbstractTile[str, Tile]):
    tile_type: TileType = TileType.CHUNK
    dimensions: tuple[int, int] | None = None
    biome: TileType | None = None
    origin: str | None = None  # what is tile id at (0,0) of the nXm dim land



import logging

from pydantic import BaseModel
from pymongo.results import DeleteResult

from ...client.dao import DaoClient
from ...sdk.contracts.dtos.entities.entity import Entity
from ...sdk.contracts.dtos.sdk.requests.chunk.create import ChunkCreateRequest
from ...sdk.contracts.dtos.sdk.requests.chunk.delete import ChunkDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.chunk.get import ChunkGetRequest
from ...sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ...sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ...sdk.contracts.dtos.tiles.address import Address
from ...sdk.contracts.dtos.tiles.chunk import Chunk
from ...sdk.contracts.dtos.tiles.tile import Tile
from ...sdk.contracts.dtos.tiles.world import World
from ...sdk.contracts.types.dao_document import DaoDocumentType
from ..factories.chunk.flat import FlatChunkFactory
from ..factories.entity.entity import EntityFactory
from ..factories.world.world import WorldFactory

logger = logging.getLogger()


class StateService(BaseModel):
    daoclient: DaoClient

    world_factory: WorldFactory
    entity_factory: EntityFactory
    flatchunk_factory: FlatChunkFactory

    class Config:
        arbitrary_types_allowed = True

    ### Worlds ##################################

    async def worlds_get(self) -> list[World]:
        return await self.daoclient.get_all(doc_type=DaoDocumentType.WORLD)

    ### World ##################################

    async def world_create(self, request: WorldCreateRequest) -> str:
        logger.debug("[StateService] creating world")
        return await self.world_factory.create(name=request.name)

    async def world_lite_get(self, request: WorldGetRequest) -> World:
        address_world: Address = Address.model_validate({"world_id": request.id})
        return await self.daoclient.get(address=address_world)

    async def world_get(self, request: WorldGetRequest) -> World:
        # TODO: needs conversion to daoclient backend

        # Build a complete World from Lite objects
        address_world: Address = Address.model_validate({"world_id": request.id})
        world: World = await self.daoclient.get(address=address_world)

        partial_world = world.model_dump(exclude={"ids"})
        world: World = World.model_validate(partial_world)

        chunk_ids: set[str] = world.ids
        for chunk_id in chunk_ids:
            local_chunk: Chunk = await self.chunk_get(request=ChunkGetRequest(world_id=request.id, chunk_id=chunk_id))
            world.contents[chunk_id] = local_chunk
        return world

    async def world_delete(self, request: WorldDeleteRequest) -> bool:
        logger.debug("[StateService] deleting world")
        address_world: Address = Address.model_validate({"world_id": request.id})
        result: DeleteResult = await self.daoclient.delete(address=address_world)
        if result.deleted_count == 1:
            return True
        return False

    ### Chunk ##################################

    async def chunk_create(self, request: ChunkCreateRequest) -> str:
        logger.debug("[StateService] creating chunk")
        new_chunk: Chunk = await self.flatchunk_factory.create(
            world_id=request.world_id, name=request.name, dimensions=request.dimensions, biome=request.biome
        )

        # Entity Factory Terrain Creation
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": new_chunk.id})
        await self.entity_factory.terrain_generate(address=address_chunk, chunk=new_chunk)

        new_chunk: Chunk = await self.daoclient.get(address=address_chunk)

        # Entity Factory Quantum
        await self.entity_factory.quantum(address=address_chunk, chunk=new_chunk)

        return new_chunk.id

    async def chunk_delete(self, request: ChunkDeleteRequest) -> None:
        msg: str = f"[WorldService] deleting chunk {id}"
        logger.debug(msg)
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})
        await self.daoclient.delete(address=address_chunk)

    async def chunk_lite_get(self, request: ChunkGetRequest) -> Chunk:
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})
        return await self.daoclient.get(address=address_chunk)

    async def chunk_get(self, request: ChunkGetRequest) -> Chunk:
        # Builds a complete Chunk from Lite objects
        address_chunk: Address = Address.model_validate({"world_id": request.world_id, "chunk_id": request.chunk_id})
        chunk: Chunk = await self.daoclient.get(address=address_chunk)

        chunk_partial = chunk.model_dump(exclude={"tile_ids"})
        chunk: Chunk = Chunk.model_validate(chunk_partial)

        # re-hydrate the tiles
        tile_ids: set[str] = chunk.ids
        for tile_id in tile_ids:
            address_tile: Address = Address.model_validate(
                {"world_id": request.world_id, "chunk_id": chunk.id, "tile_id": tile_id}
            )
            tile: Tile = await self.tile_get(address=address_tile)

            # re-hydrate the entities
            entity_ids: set[str] = tile.ids
            for entity_id in entity_ids:
                address_entity: Address = Address.model_validate(
                    {
                        "world_id": request.world_id,
                        "chunk_id": chunk.id,
                        "tile_id": tile_id,
                        "entity_id": entity_id,
                    }
                )
                entity: Entity = await self.entity_get(address=address_entity)

                # add finalized entity to tile
                tile.contents[entity_id] = entity

            # Add finalized tile to chunk
            chunk.contents[tile_id] = tile

        return chunk

    ### Tile ##################################

    async def tile_get(self, address: Address) -> Tile:
        return await self.daoclient.get(address=address)

    ### Entity ##################################

    async def entity_get(self, address: Address) -> Entity:
        return await self.daoclient.get(address=address)



from enum import Enum


class EntityType(str, Enum):
    UNKNOWN = "unknown"

    GRASS = "grass"
    TREE = "tree"
    FRUIT = "fruit"

    FISH = "fish"

    MYCELIUM = "mycelium"
    MUSHROOM = "mushroom"



import logging
import traceback

from fastapi import APIRouter, HTTPException

from .....sdk.contracts.dtos.sdk.responses.response import Response
from .....sdk.contracts.dtos.sdk.responses.worlds.get import WorldsGetResponse
from .....sdk.contracts.dtos.tiles.world import World
from .....sdk.contracts.errors.server.dao.conflict import DaoConflictError
from .....sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from .....sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError
from ..context import ContextManager

logger = logging.getLogger()

router: APIRouter = APIRouter(
    prefix="/worlds",
    tags=["worlds"],
)


### /worlds


@router.get("")
async def worlds_get() -> Response[WorldsGetResponse]:
    try:
        worlds: list[World] = await ContextManager.state_service.worlds_get()
        response: WorldsGetResponse = WorldsGetResponse(worlds=worlds)
        return Response[WorldsGetResponse](data=response)
    except DaoConflictError as error:
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        logger.error(str(error))
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        # catch everything else
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error
